package com.es.ai;


import java.util.Arrays;

import org.apache.commons.configuration.Configuration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.es.ArraySet;
import com.es.Board;
import com.es.Board.State;
import com.es.CmdConfiguration;
import com.es.IllegalMoveException;
import com.es.pieces.Pawn;
import com.es.pieces.Piece;
import com.es.pieces.Piece.Color;

public class AlphaBetaAI {
    private static final Logger LOG = LoggerFactory.getLogger(AlphaBetaAI.class);

    private final TranspositionTable transpositionTable = new TranspositionTable();
    private int transHit = 0;
    private final Color colorPlaying;
    private final Configuration configuration;
    private final int depth;
    private final Board board;

    public AlphaBetaAI(Color colorPlaying, Board board, Configuration configuration) {
        this.colorPlaying = colorPlaying;
        this.board = board;
        this.configuration = configuration;
        this.depth = configuration.getInt(CmdConfiguration.DEPTH);
    }

    public AlphaBetaAI(Color colorPlaying, Board board) {;
        this.colorPlaying = colorPlaying;
        this.board = board;
        this.configuration = null;
        this.depth = 4;
    }

    public int getTransHit() {
        return transHit;
    }

    public static Color swapColor(Color color) {
        return color.equals(Color.WHITE) ? Color.BLACK : Color.WHITE;
    }

    public int computeNextMove(MoveNode node, Color color) {
        for(int d=2; d <= 2; d++) {
            transHit = 0;
            long start = System.currentTimeMillis();
            alphabeta(node, d, -1000000, 1000000, color);
            
            try {
                board.checkBoard();
            } catch (IllegalMoveException e) {
                LOG.error("Error with board: {}", e.getMessage());
            }
            
            long time = System.currentTimeMillis() - start;

//            System.out.println("DEPTH: " + d + " TT HITS: " + transHit + " TIME: " + time + " NODES: " + node.getNodeCount() + " CHILD: " + node.getChildCount());
//            System.out.println(node.childrenToString());
        }

        return node.getBestChild().getMove();
    }

    public int alphabeta(MoveNode node, int depth, int alpha, int beta, Color color) {
        if(depth == 0) {
            int score = computeScore();
            node.setScore(score);
            node.setDepth(depth);
            node.setRetVal(score);
            return score;
        }

        int[] boardPieces = board.getPieces(color);
        final int[] childrenPieces = node.getChildrenPieces();

/*        
        // we want to make sure we walk through the pieces in the same order as any children
        if(childrenPieces.length < boardPieces.length && childrenPieces.length != 0) {
            final int[] boardNoChildrenPieces = Arrays.copyOf(boardPieces, boardPieces.length);

            // remove all the children from the board
            for(int p:childrenPieces) {
                ArraySet.removeNumber(boardNoChildrenPieces, p, Board.MAX_SQUARE);
            }

            boardPieces = new int[boardPieces.length];
            int i=0;

            // copy over all of the children pieces first
            for(int childPiece:childrenPieces) {
                // copy over the child piece, only if it's still on the board
                if(Arrays.binarySearch(board.getPieces(color), childPiece) >= 0)
                    boardPieces[i++] = childPiece;
            }

            final int length = i;

            // copy over all the rest of the pieces from the board
            for( ; boardNoChildrenPieces[i - length] != Board.MAX_SQUARE; ++i) {
                boardPieces[i] = boardNoChildrenPieces[i - length];
            }

            // ensure any other spaces are set to MAX_SQUARE
            for( ; i < boardPieces.length; ++i) {
                boardPieces[i] = Board.MAX_SQUARE;
            }
        }
*/

        // TODO: Remove this check
        try {
            board.checkBoard();
        } catch (IllegalMoveException e) {
            LOG.error("Error with board: " + e.getMessage());
        }

        int[] allMoves = this.generateAllMoves(boardPieces);
        int[] ret = { 0, alpha, beta };

        for(int i = 0; i < allMoves.length && Board.getFromSquare(allMoves[i]) != Board.MAX_SQUARE; ++i) {

            MoveNode child = node.findChild(allMoves[i]);

            if(child != null) {
                int score = alphabeta(child, depth - 1, alpha, beta, swapColor(color));

                // get the alpha or beta depending upon who's turn it is
                if(colorPlaying.equals(color)) {
                    alpha = Math.max(alpha, score);
                } else {
                    beta = Math.min(beta, score);
                }

            } else {
                // compute alpha-beta for the move
                ret = alphabeta(node, depth, allMoves[i], alpha, beta, color);

                // update the values of alpha and beta
                alpha = ret[0];
                beta = ret[1];
            }

            if(beta <= alpha) {
                // node.removeChildrenAfter(child);
                node.removeNotAtDepth(depth-1);
                break;
            }
        }

        final int retVal = colorPlaying.equals(color) ? alpha : beta;

        int score = retVal;

        if(node.getChildCount() != 0) {
            score = (colorPlaying.equals(color) ? node.getBestChild() : node.getWorstChild()).getScore();
        }

        node.setScore(score);
        node.setDepth(depth);
        node.setRetVal(retVal);

        return retVal;
    }


    public int[] alphabeta(MoveNode node, int depth, int move, int alpha, int beta, Color color) {
        State state;
        
        try {
            state = board.makeMove(move);
        } catch (IllegalMoveException e) {
            if(!e.isKingInCheck()) {
                LOG.error("Illegal move during compute: {}", e.getMessage());
                System.out.println(board.toString());
                System.exit(-1);
            }

            // return here, but continue searching
            return new int[] { alpha, beta };
        }

        MoveNode childNode = transpositionTable.get(board);

        // check to see if we need to create a new node, or if we can use the one from the table
        if(childNode == null || childNode.getDepth() <= depth) {
            childNode = new MoveNode(node, move);

            final int score = alphabeta(childNode, depth - 1, alpha, beta, swapColor(color));

            // get the alpha or beta depending upon who's turn it is
            if(colorPlaying.equals(color)) {
                alpha = Math.max(alpha, score);
            } else {
                beta = Math.min(beta, score);
            }

            // add the new node
            node.addChild(childNode);

            // add it to the transposition table
            transpositionTable.put(board, childNode);
        } else {
            if(colorPlaying.equals(color)) {
                alpha = Math.max(alpha, childNode.getRetVal());
            } else {
                beta = Math.min(beta, childNode.getRetVal());
            }

            transHit++;

            // add the child node's children to the current node's children
            node.addChildren(childNode);
        }
        
        // unmake the move
        try {
            board.unmakeMove(move, state);
        } catch (IllegalMoveException e) {
            LOG.error("Illegal move during compute: {}", e.getMessage());
            System.out.println(board.toString());
            System.exit(-1);
        }

        // see if we have a cut-off
        if(beta <= alpha) {
            node.setDepth(depth);
            node.setRetVal(colorPlaying.equals(color) ? alpha : beta);
            if(node.getChildCount() != 0) {
                node.setScore((colorPlaying.equals(color) ? node.getBestChild() : node.getWorstChild()).getScore());
            } else {
                node.setScore(node.getRetVal());
            }

            return new int[] {alpha, beta };
        }

        return new int[] { alpha, beta };
    }

    public int[] generateAllMoves(int[] pieces) {
        int[] allMoves = new int[161];
        int i = 0;

        for(int p:pieces) {
            if(p == Board.MAX_SQUARE) {
                break;  // in sorted order, so we can break early
            }
            
            final Piece piece = board.getPiece(p);
            int[] moves = piece.generateAllMoves(board, p);

            for(int m:moves) {
                if(m == Board.MAX_SQUARE) {
                    break;  // always in sorted order, so we're done here
                }
                
                // check to see if we have a pawn promoting
                if(piece instanceof Pawn && ( (m & 0xf0) == 0x70 || (m & 0xf0) == 0x00) ) {
                    allMoves[i++] = Board.createMoveValue(p, m, 'q');
                    allMoves[i++] = Board.createMoveValue(p, m, 'b');
                    allMoves[i++] = Board.createMoveValue(p, m, 'n');
                    allMoves[i++] = Board.createMoveValue(p, m, 'r');
                } else {
                    allMoves[i++] = Board.createMoveValue(p, m, '-');
                }
            }
        }

        Arrays.fill(allMoves, i, allMoves.length, Board.createMoveValue(Board.MAX_SQUARE, Board.MAX_SQUARE, '-'));
        return allMoves;
    }

    public int computeScore() {
        final int[] whitePieces = board.getPieces(Color.WHITE);
        final int[] blackPieces = board.getPieces(Color.BLACK);
        int whiteScore = 0;
        int blackScore = 0;

        //
        // Compute the value based upon position
        //
        for(int p:whitePieces) {
            if(p == Board.MAX_SQUARE) {
                break;
            }
            whiteScore += board.getPiece(p).getPositionValue(p);
        }

        for(int p:blackPieces) {
            if(p == Board.MAX_SQUARE) {
                break;
            }
            blackScore += board.getPiece(p).getPositionValue(p);
        }

        //
        // Add in bonus for each piece attacking and defending
        //
        whiteScore += computeAttackDefendBonus(board, whitePieces, Color.BLACK);
        blackScore += computeAttackDefendBonus(board, blackPieces, Color.WHITE);


        if(LOG.isDebugEnabled()) {
//            LOG.info("MOVE: {} -> {}", Integer.toHexString(node.getMove()[0]), Integer.toHexString(node.getMove()[1]));
//            LOG.info("WHITE: {} BLACK: {} SCORE: " + (colorPlaying.equals(Color.WHITE) ? whiteScore - blackScore : blackScore - whiteScore), whiteScore, blackScore);
        }

        return colorPlaying.equals(Color.WHITE) ? whiteScore - blackScore : blackScore - whiteScore;
    }

    public int computeAttackDefendBonus(Board board, int[] pieces, Color targetColor) {
        int ret = 0;

        for(int p:pieces) {
            if(p == Board.MAX_SQUARE) {
                break;
            }

            final Piece piece = board.getPiece(p);
            final int[] moves = piece.generateAllMoves(board, p);

            for(int m:moves) {
                if(m == Board.MAX_SQUARE) {
                    break;
                }

                final Piece targetPiece = board.getPiece(m);

                if(targetPiece == null) {
                    continue;
                }

                if(targetPiece.getColor().equals(targetColor)) {
                    // check for an attack
                    ret += targetPiece.getValue() * .5;
                } else {
                    // check for a defense
                    ret += targetPiece.getValue() * 0.250;
                }
            }
        }

        return ret;
    }

}
